
import numpy as np
import csv
import math

def size_of_tree( matrix ):
    return np.sum( np.absolute( matrix ) ) - matrix.shape[1]

def create_stage( curr_idx, idxs ):
    # group by 2 and create ops
    op_list = []
    for i in range( int( math.ceil( len(idxs) / 2 ) ) ):
        a = idxs[2*i]
        if len(idxs) < 2*i+2:
            b = [ -1, 0, False ]
        else:
            b = idxs[2*i + 1]
        add_op = 4*a[2] + 2*b[2]
        op_new = [ curr_idx, a[0], b[0], -1, add_op, 0, 0, 0 ]
        curr_idx += 1
        op_list += [ op_new ]
    return op_list, curr_idx

def create_stage_ternary( curr_idx, idxs ):
    # group by 3 and create ops
    op_list = []
    for i in range( int( math.ceil( len(idxs) / 3 ) ) ):
        a = idxs[3*i]
        if len(idxs) < 3*i+2:
            b = [ -1, 0, False ]
        else:
            b = idxs[3*i + 1]
        if len(idxs) < 3*i+3:
            c = [ -1, 0, False ]
        else:
            c = idxs[3*i + 2]
        add_op = 4*a[2] + 2*b[2] + 1*c[2]
        op_new = [ curr_idx, a[0], b[0], c[0], add_op, 0, 0, 0 ]
        curr_idx += 1
        op_list += [ op_new ]
    return op_list, curr_idx

def create_ops_for_tree( curr_idx, curr_idxs, use_tern = False ):
    # curr_idxs = [ ( idx, depth_avail, is_pos ) ]
    if len(curr_idxs) == 0:
        return ( [], curr_idx, -1, 0 )
    curr_d = 0
    op_list = []
    reserves = []
    while len(reserves) > 0 or len(curr_idxs) > 1 or len(op_list) < 1:
        reserves = [ x for x in curr_idxs if x[1] > curr_d ]
        to_reduce = [ x for x in curr_idxs if x[1] <= curr_d ]
        if use_tern:
            reduced_ops, curr_idx = create_stage_ternary( curr_idx, to_reduce )
        else:
            reduced_ops, curr_idx = create_stage( curr_idx, to_reduce )
        curr_idxs = [ ( x[0], curr_d + 1, True )  for x in reduced_ops ] + reserves
        op_list += reduced_ops
        curr_d += 1
    output_idx = op_list[-1][0]
    return op_list, curr_idx, output_idx, curr_d

def combine_dep( dep_a, dep_b ):
    # merge two rows back together
    for x in dep_b:
        dep_a[x] = dep_b[x]
    return dep_a

def get_dependancies( idxs, matrix, no_in, no_out ):
    # merge certain indexes back together
    dependancies = {}
    for i in idxs:
        dependancies[ i ] = list( np.nonzero( matrix[no_in:,i] )[0] + no_out )
        dependancies = combine_dep( dependancies,
                                    get_dependancies( dependancies[ i ], matrix, no_in, no_out )
        )
    return dependancies

def make_tree( matrix, no_in, no_out ):
    '''
    Convert the matrix to the op list
    '''
    dependancies = {}
    output_depths = {}
    for j in range( no_out ):
        idxs = list( np.nonzero( matrix[no_in:,j] )[0] + no_out )
        dependancies[j] = idxs
        dependancies = combine_dep( dependancies,
                                    get_dependancies( idxs, matrix, no_in, no_out ) )
    outputs = {}
    op_list = []
    op_idx = no_in
    max_depth = 0
    while len( outputs ) < len(dependancies):
        for x in dependancies:
            if x not in outputs: # haven't already done
                # check all dependancies are resolved otherwise skip for now
                dep_depths = [ output_depths[ y ] for y in dependancies[x] if y in output_depths ]
                if len( dep_depths ) == len(dependancies[x]):
                    idxs = np.nonzero( matrix[:,x] )[0]
                    idxs_in = []
                    for i in idxs:
                        j = i
                        d = 0
                        if i >= no_in:
                            j = outputs[ i + no_out - no_in ]
                            d = output_depths[ i + no_out - no_in ]
                        assert matrix[i,x] != 0, "has dependancy with zero? - invalid matrix"
                        idxs_in += [ (j, d, matrix[i,x] > 0) ]
                    new_ops, op_idx, output_idx, curr_d = create_ops_for_tree( op_idx, idxs_in )
                    outputs[ x ] = output_idx
                    output_depths[ x ] = curr_d
                    if curr_d > max_depth:
                        max_depth = curr_d
                    op_list += new_ops
    return op_list, [ outputs[i] for i in range( no_out ) ], max_depth

def reverse_check_result( orig_mat, new_mat ):
    '''
    A verification step that the tree is valid
    '''
    no_in = orig_mat.shape[1]
    no_out = orig_mat.shape[0]
    # replace elim outputs with full expression
    for i in range( new_mat.shape[0] - no_out ):
        vec_idx_orig = np.nonzero( new_mat[no_out+i,:] )[0]
        update_idx = np.nonzero( new_mat[:,no_in+i] )[0]
        vec_idx = np.tile( vec_idx_orig, len(update_idx) )
        update_idx = np.repeat( update_idx, len(vec_idx_orig) )
        vec = new_mat[no_out+i,vec_idx]
        new_mat[update_idx,vec_idx] += vec*new_mat[update_idx,no_in+i]
        new_mat[no_out+i,vec_idx_orig] -= new_mat[no_out+i,vec_idx_orig]
    return np.sum( new_mat[:no_out,:no_in] == orig_mat ) == no_in*no_out

def get_matrix( fname ):
    '''
    Load the matrix
    '''
    f = open( fname )
    rdr = csv.reader( f )
    data = [ [ int(y) for y in x ] for x in rdr ]
    matrix = np.transpose( np.array( data, dtype = np.int16 ) )
    no_in = matrix.shape[1]
    no_out = matrix.shape[0]
    print( "no_in = " + str(no_in) + ", no_out = " + str(no_out) )
    initial_no_adds = size_of_tree( matrix )
    print( "initial matrix is " + str( initial_no_adds  ) )
    return matrix, no_in, no_out, initial_no_adds

def write_output( fname, matrix, initial_no_adds, no_in, no_out, remove_neg_ops = True ):
    '''
    Write the CSE result to a file
    '''
    final_no_adds = size_of_tree( matrix )
    print( "improvement is from " + str( initial_no_adds ) + " to " +
           str( final_no_adds ) + " or " + str( final_no_adds*100/initial_no_adds ) + "%" )
    f_out = open( fname, "w" )
    wrt = csv.writer( f_out )
    tree_ops, outputs, max_depth = make_tree( np.transpose( matrix ), no_in, no_out )
    tmp = wrt.writerow( outputs )
    for x in tree_ops:
        tmp = wrt.writerow( x )
    f_out.close()
    if remove_neg_ops:
        remove_negate_ops( fname )
    return max_depth

def compute_op( a, b, c, op_code ):
    if op_code == 0:
      return - a - b - c
    if op_code == 1:
      return - a - b + c
    if op_code == 2:
      return - a + b - c
    if op_code == 3:
      return - a + b + c
    if op_code == 4:
      return a - b - c
    if op_code == 5:
      return a - b + c
    if op_code == 6:
      return a + b - c
    return a + b + c

def remove_negate_ops( fname ):
    '''
    Move any ops that have double negate
    This is necessary for serial adders as it isn't hardware friendy to implement
    '''
    f = open( fname )
    rdr = csv.reader( f )
    data = [ x for x in rdr ]
    f.close()
    f_out = open( fname, "w" )
    wrt = csv.writer( f_out )
    negate_ops = set([])
    output_idxs = data[0]
    output_idxs_set = set( [ int(x) for x in output_idxs ] )
    negate_outputs = []
    new_rows = []
    last_op = -1
    for x in data[1:]:
        new_row = x
        op_code = int( x[4] )
        if int( x[1] ) in negate_ops:
            op_code = ( op_code + 4 ) % 8
        if int( x[2] ) in negate_ops:
            op_code = ( ( op_code + 2 ) % 4 ) + 4*( op_code >> 2 )
        if int( x[3] ) in negate_ops:
            op_code = ( ( op_code + 1 ) % 2 ) + 2*( op_code >> 1 )
        new_row[4] = str( op_code )
        is_neg = op_code == 0 or ( int(x[3]) < 0 and op_code < 2 )
        if is_neg:
            negate_ops.add( int(x[0]) )
            new_row[4] = str( 7 - op_code )
            if int( x[0] ) in output_idxs_set:
                negate_outputs += [ int(x[0]) ]
        new_rows += [ new_row ]
        last_op = int(new_row[0])
    # add more regs that fix negated outputs
    for i in negate_outputs:
        last_op += 1
        new_row = [ last_op, i, -1, -1, 0, 0, 0, 0 ]
        for j, oi in enumerate( output_idxs ):
            if oi == str(i):
                output_idxs[j] = str(last_op)
                break
        new_rows += [ new_row ]
    wrt.writerow( output_idxs )
    for new_row in new_rows:
        wrt.writerow( new_row )
    f_out.close()

def verify_tree( fname, fname_out ):
    '''
    Verify that the op list generated matches the matrix input
    '''
    matrix, no_in, no_out, initial_no_adds = get_matrix( fname )
    f_t = open( fname_out )
    rdr = csv.reader( f_t )
    ops = [ [ int(y) for y in x ] for x in rdr ]
    outputs = ops[0]
    tmp_inputs = np.random.randint( -(1<<12), 1 << 12, no_in )
    expected_out = np.matmul( matrix, tmp_inputs )
    tmp_vals = {}
    for op in ops[1:]:
        a = tmp_inputs[op[1]] if op[1] not in tmp_vals else tmp_vals[op[1]]
        if op[2] >= 0:
            b = tmp_inputs[op[2]] if op[2] not in tmp_vals else tmp_vals[op[2]]
        else:
            b = 0
        if op[3] >= 0:
            c = tmp_inputs[op[3]] if op[3] not in tmp_vals else tmp_vals[op[3]]
        else:
            c = 0
        tmp_vals[op[0]] = compute_op( a, b, c, op[4] )
    for i, o in enumerate( outputs ):
        if o < 0:
            continue
        assert expected_out[i] == tmp_vals[o], "must match expected output from tree for output " + str(o)
    return True




def create_serial_add_op( names, op_code, shifts, BW_in, BW_out, module_name, reset_name, depth ):
    '''
    Creates an add or subtract operation ( input c is not allowed, no shifts allowed )
    names = [ out_name, a_name, b_name, c_name ]
    op_code = 0 -> - a - b - c ( not allowed in this implementation )
    op_code = 1 -> - a - b + c ( not allowed in this implementation )
    op_code = 2 -> - a + b - c
    op_code = 3 -> - a + b + c
    op_code = 4 ->   a - b - c
    op_code = 5 ->   a - b + c
    op_code = 6 ->   a + b - c
    op_code = 7 ->   a + b + c
    shifts = [ shift_a, shift_b, shift_c ] ( not allowed in this implementation )
    BW_in -> the bitwidth of the input
    BW_out -> the bitwidth of the output
    module_name -> the name to call the adder
    reset_name -> the reset to use, will assert the reset on the cycle before use
    depth -> how deep the op is down the tree, inputs have a depth of 0 ( not used in this op )
    '''
    assert names[3] == str(BW_out) + "'h0", "Three inputs not implemented in this adder"
    assert shifts[0] == 0 and shifts[1] == 0, "Shift registers not implemented in this adder"
    # deal with neg regs
    op_code = op_code >> 1
    if names[2] == str(BW_out) + "'h0" and op_code == 0:
        op_code = 1;
    assert op_code != 0, "Cannot be both negative inputs for serial adder for module " + module_name
    neg_b = ( op_code != 3 )*1
    # swap the order so input a is positive
    if op_code == 1:
        names[1], names[2] = ( names[2], names[1] )
    ser_add_module = '''wire [''' + str( BW_in - 1 ) + ":0] " + names[0] + ''';
serial_adder #(
  .neg_b(''' + str(neg_b) + '''),
  .BW(''' + str(BW_in) + ''')
) ''' + module_name + ''' (
 .clk(clock),
 .reset(''' + reset_name + '''),
 .a(''' + names[1] + '''),
 .b(''' + names[2] + '''),
 .c(''' + names[0] + ''')
);
'''
    return ser_add_module, BW_in

def write_serial_adder_module( fname ):
    f = open( fname, "w" )
    f.write( """
`timescale 1ns / 1ps

module serial_adder
#(
 parameter neg_b = 0,
 parameter BW = 16
)
(
 input 		 clk,
 input 		 reset,
 input [BW-1:0]  a,
 input [BW-1:0]  b,
 output [BW-1:0] c
 );
   reg carry_reg;
   reg [BW-1:0] data_out_reg;
   wire [BW-1:0] b_neg;
   wire [BW:0] 	 result;
   assign b_neg = neg_b ? ~b : b;
   assign result = a + b_neg + carry_reg;
   assign c = data_out_reg;
always @( posedge clk ) begin
   if ( reset ) begin
      carry_reg <= neg_b;
   end else begin
      carry_reg <= result[BW];
   end
   data_out_reg <= result[BW-1:0];
end
endmodule
""" )
    f.close()

def create_normal_add_op( names, op_code, shifts, BW_in, BW_out, module_name, reset_name, depth ):
    '''
    Creates an add or subtract operation
    names = [ out_name, a_name, b_name, c_name ]
    op_code = 0 -> - a - b - c
    op_code = 1 -> - a - b + c
    op_code = 2 -> - a + b - c
    op_code = 3 -> - a + b + c
    op_code = 4 ->   a - b - c
    op_code = 5 ->   a - b + c
    op_code = 6 ->   a + b - c
    op_code = 7 ->   a + b + c
    shifts = [ shift_a, shift_b, shift_c ]
    BW_in -> the bitwidth of the input
    BW_out -> the bitwidth of the output
    module_name -> the name to call the adder
    reset_name -> the reset to use, will assert the reset on the cycle before use ( not used in this op )
    depth -> how deep the op is down the tree, inputs have a depth of 0 ( not used in this op )
    '''
    op_str = ""
    op_mod = 4
    for n, s in zip( names[1:], shifts ):
        if op_code < op_mod:
            op_str += " - "
        elif op_mod != 4:
            op_str += " + "
        op_code = op_code % op_mod
        op_mod = op_mod >> 1
        op_str += "( $signed( " + n + " ) "
        if s < 0:
            op_str += " >> " + str( - s ) + " )"
        elif s > 0:
            op_str += " << " + str( s ) + " )"
        else:
            op_str += ")"
    add_op = '''reg [''' + str( BW_in - 1 ) + ":0] " + names[0] + ''';
always @( posedge clock ) begin
''' + names[0] + " <= " + op_str + ''';
end
'''
    return add_op, BW_in




from .bu_CSE import bu_CSE
from .td_CSE import td_CSE
from .cse_common import get_matrix
from .cse_common import write_output
from .cse_common import verify_tree
from .sparse_mat_mul_generator import SMM_generate
from .add_ops import create_serial_add_op
from .add_ops import create_normal_add_op
from .add_ops import write_serial_adder_module
from .helper_compute import round_to
from .helper_compute import floor_to
from .helper_compute import get_ternary
from .helper_compute import conv2d
from .helper_compute import conv1d
from .helper_compute import get_AB
from .helper_compute import maxpool2d
from .helper_compute import maxpool1d
from .helper_compute import relu
from .convert_tree_to_c import write_matrix_to_c_ary
from .convert_tree_to_c import write_bn_relu_to_c
from .convert_tree_to_c import write_tree_to_c

__name__ = "twn_generator"
__version__ = "0.1.2"




import numpy as np
import math
import sys
from . import cse_common

def get_idx( i, j, no_in ):
    assert i != j and i < j, "Invalid index"
    return int(i*no_in - i*(i+1)/2 + j - i - 1)

def idx_to_ij( idx, no_in ):
    i = 0
    for i in range(no_in):
        if idx < i*no_in - i*(i+1)/2:
            break
    i = i - 1
    j = int(idx - i*no_in + i*(i+1)/2 + i + 1)
    return i, j

def copy_across( same_in, diff_in, no_in ):
    same_out = np.ones( int(no_in * ( no_in - 1 ) / 2), dtype=np.int16 )
    diff_out = np.ones( int(no_in * ( no_in - 1 ) / 2), dtype=np.int16 )
    for i in range( 0, no_in - 1 ):
        start_out = get_idx( i, i + 1, no_in )
        stop_out = get_idx( i, no_in - 1, no_in )
        start_in = get_idx( i, i + 1, no_in - 1 )
        stop_in = get_idx( i, no_in - 1, no_in - 1 )
        same_out[start_out:stop_out] = same_in[start_in:stop_in]
        diff_out[start_out:stop_out] = diff_in[start_in:stop_in]
    return same_out, diff_out

# find the most common subexpression with 2 inputs
def find_common_2( matrix, update_idxs = [], prev_cnts_same = None, prev_cnts_diff = None ):
    no_in = matrix.shape[1]
    no_out = matrix.shape[0]
    if prev_cnts_same is not None and prev_cnts_diff is not None:
        # fill from prev to avoid recounting
        counts_same, counts_diff = copy_across( prev_cnts_same, prev_cnts_diff, no_in )
    else:
        counts_same = np.ones( int(no_in * ( no_in - 1 ) / 2), dtype=np.int16 )
        counts_diff = np.ones( int(no_in * ( no_in - 1 ) / 2), dtype=np.int16 )
    # build the combinations to compute
    col_1_idxs = []
    col_2_idxs = []
    idxs = []
    for i in update_idxs:
        col_1_idx = list(range(0,i)) + list(range(i + 1,no_in))
        for j in update_idxs:
            if i == j:
                break
            else:
                col_1_idx.remove(j)
        # can filter out values that have zero. can only decrease the number in common ( except last )
        col_2_idx = [i]*len(col_1_idx)
        idx = [ get_idx( i, j, no_in ) if i < j else get_idx( j, i, no_in ) for i, j in zip(col_1_idx, col_2_idx) ]
        idxs += idx
        col_1_idxs += col_1_idx
        col_2_idxs += col_2_idx
    # look at the computed indexes and find the final
    idxs = np.array(idxs, dtype = np.int32 )
    filt_idx = ( counts_same[idxs] + counts_diff[idxs] ) > 0
    idxs = idxs[filt_idx]
    col_1_idxs = np.array( col_1_idxs )[filt_idx]
    col_2_idxs = np.array( col_2_idxs )[filt_idx]
    col_1 = matrix[:,col_1_idxs]
    col_2 = matrix[:,col_2_idxs]
    same_cnt = np.sum( ( col_1 == col_2 ) * np.abs(col_1), 0 )  # either (1,1) or (-1,-1)
    diff_cnt = np.sum( ( col_1 == -col_2 ) * np.abs(col_1), 0 ) # either (1,-1) or (-1,1)
    counts_same[idxs] = same_cnt
    counts_diff[idxs] = diff_cnt
    print( "updated " + str(len(idxs)) + " values, skipped " + str(len(filt_idx) - len(idxs)) )
    # determine the most common subexpression
    max_idx_same = np.argmax( counts_same )
    max_idx_diff = np.argmax( counts_diff )
    max_idx = max_idx_diff
    same = False
    counts = counts_diff
    if max_idx_same >= max_idx_diff:
        max_idx = max_idx_same
        same = True
        counts = counts_same
    n, m = idx_to_ij( max_idx, no_in ) # n and m are the indexs of the inputs for the most common subexpression
    # get the indexs of matrix rows that contain this subexpression
    if same:
        pos_idxs = np.arange(no_out)[ ( np.abs( matrix[:,n] + matrix[:,m] ) == 2 ) * (  matrix[:,n] > 0 ) ]
        neg_idxs = np.arange(no_out)[ ( np.abs( matrix[:,n] + matrix[:,m] ) == 2 ) * (  matrix[:,n] < 0 ) ]
    else:
        pos_idxs = np.arange(no_out)[ ( np.abs( matrix[:,n] - matrix[:,m] ) == 2 ) * (  matrix[:,n] > 0 ) ]
        neg_idxs = np.arange(no_out)[ ( np.abs( matrix[:,n] - matrix[:,m] ) == 2 ) * (  matrix[:,n] < 0 ) ]
    return n, m, pos_idxs, neg_idxs, same, counts_same, counts_diff

def td_CSE( matrix ):
    orig_mat = matrix.copy() # copy for later verification
    new_rows = [] # to store removed subexpressions
    # nothing is precomputed so update everything
    cnts_same = None
    cnts_diff = None
    update_idxs = list(range(matrix.shape[1]))
    while True:
        n, m, pos_idxs, neg_idxs, same, cnts_same, cnts_diff = find_common_2( matrix, update_idxs, cnts_same, cnts_diff )
        if len(pos_idxs) + len(neg_idxs) <= 1:
            break
        update_idxs = [ n, m, matrix.shape[1] ]
        pattern = np.zeros( matrix.shape[1], dtype = np.int16 )
        pattern[n] = 1
        if same:
            pattern[m] = 1
        else:
            pattern[m] = -1
        for i in pos_idxs:
            matrix[i,:] = matrix[i,:] - pattern
        for i in neg_idxs:
            matrix[i,:] = matrix[i,:] + pattern
        matrix = np.hstack( [ matrix, np.zeros( ( matrix.shape[0], 1 ), dtype = np.int16 ) ] )
        for i in pos_idxs:
            matrix[i,-1] = 1
        for i in neg_idxs:
            matrix[i,-1] = -1
        new_rows += [ pattern ]
    print( "Finalizing matrix ... " )
    # put the removed subexpressions back into the matrix
    for i, x in enumerate( new_rows ):
        tmp = np.concatenate( ( np.array( x, dtype=np.int16 ),  np.zeros( ( matrix.shape[1] - len(x) ), dtype = np.int16 ) ) )
        matrix = np.vstack( [ matrix, tmp ] )
    print( "Checking ... " )
    assert cse_common.reverse_check_result( orig_mat, matrix.copy() ), "must have same matrix originally"
    print( "Verifed correctness" )
    return matrix

